/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northeastern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com)
 */

#include <numeric>
#include <sstream>
#include <string>
#include <memory>

#include "SLTKUtility.h"
#include "SLTKDataUtility.h"
#include "../../tensor/core/CHeader.h"

namespace util
{

/* check if the file exists */
bool IsFileExists(const char *fileName)
{
    CALL;

    std::ifstream infile(fileName);

    OK;
    return infile.good();
}

/* get the size of a file */
std::ifstream::pos_type FileSize(const char *filename)
{
    CALL;
    std::ifstream in(filename, std::ifstream::ate | std::ifstream::binary);
    OK;
    return in.tellg();
}

/*
load word2vec from file
>>> word2vecFile - the word2vec file
>>> embSize - the input size of a single word embedding
*/
Word2Vec::Word2Vec(const char *word2VecFile, int embSize)
{
    CALL;

    /*
     * format of word2vec file:
     * 1st line: <vocab size>,<embedding size>
     * 2nd~nth line: <token> <embedding> <embedding> ...
     * todo: speedup the read operation
     */
    wifstream file(word2VecFile);
    CheckNTErrors(file.is_open(), "failed to open the embeddings file!");

    file.imbue(utf8_locale);
    wstring buffer((istreambuf_iterator<wchar_t>(file)), istreambuf_iterator<wchar_t>());
    buffer += L'\n';
    int pos = 0;
    wstring word;

    /* get vocab size */
    while(buffer[pos] != L' ')
        word += buffer[pos++];
    const int vocabSize = stoi(word);
    word.clear();

    /* get embedding size */
    while(buffer[pos] != L'\n')
        word += buffer[pos++];
    CheckNTErrors(embSize == stoi(word), "embedding size is unmatched!");
    pos++;
    word.clear();

    DTYPE *w2v = DBG_NEW DTYPE[vocabSize * embSize];
    int number(0);

    /* load embeddings for each vocab */
    for(int i = 0; i < vocabSize; ++i){

        /* load token */
        while(buffer[pos] != L' ')
            word += buffer[pos++];
        word2Id[word] = number;
        id2Word[number++] = word;
        pos++;
        word.clear();

        /* load embedding */
        for(int j = 0; j < embSize; ++j){
            while(buffer[pos] != L'\n' && buffer[pos] != L' '){
                word += buffer[pos++];
            }
            *w2v++ = stof(word);
            pos++;
            word.clear();
        }
    }
    InitTensor2D(&vec, vocabSize, embSize);
    w2v -= vocabSize * embSize;
    vec.SetData(w2v, vocabSize * embSize);
    delete[] w2v;

    OK;
}

/*
set embeddings for a mini-batch
>>> sentence - input sequence
>>> word2vec - distributed representations of words
>>> embeddings - a [batch, embSize] tensor for representing words
use random embedding if OOV
use zero embedding if PAD
use pre-trained embedding else
*/
void Word2Vec::EmbeddingLookup(XTensor * embeddings, const vector<wstring> &sentence)
{
    CALL;
    CheckNTErrors(vec.GetSize() > 1, "Uninitialized word2vec struct!");

    const int embSize = vec.GetDim(1);
    embeddings->SetZeroAll();

    int dimSize[2] = {1, embSize};
    XTensor *embedding = NewTensorBuf(2, dimSize, X_FLOAT, 1.0F, embeddings->devID, embeddings->mem);

    int i(0);
    for(const auto &word : sentence){

        /* if the input is PAD, then return a zero embedding */
        if(word == L"@"){
            _SetDataFixedFloat(embedding, 0.0F);
        }

        /* else if the input is OOV, then return a random embedding */
        else if(word2Id.find(word) == word2Id.end()){
            LOG("%ls is <UNK>, use random embedding", word.c_str());
            _SetDataFanInOut(embedding);
        }

        else{
            const int id = word2Id.at(word);
            _SelectRange(&vec, embedding, 0, id, id + 1);
        }

        int srcIndex[] = {0};
        int targetIndex[] = {i++};
        _CopyIndexed(embedding, embeddings, 0, srcIndex, 1, targetIndex, 1);
    }
    DelTensorBuf(embedding);

    OK;
}

/*
sort and pad sentences
we use the maximum length in a mini-batch for padding
*/
void BatchManager::SortAndPad()
{
    CALL;

    for(size_t i = 0; i < allSentences.size(); ++i){
        size_t maxTime(0);
        for(const auto &j : allSentences.at(i)){
            if(j.size() >= maxTime)
                maxTime = j.size();
        }
        for(size_t j = 0; j < allSentences.at(i).size(); ++j){
            for(size_t pad = allSentences.at(i).at(j).size(); pad < maxTime; ++pad){

                /* use '@' for characters padding */
                allSentences.at(i).at(j).push_back(L"@");

                /* use 0 for tags padding */
                allTags.at(i).at(j).push_back(0);
            }
        }
    }

    OK;
}

/*
index all words and tags, then dump the mappings to file
>>> fileIn - the path of input data
>>> fileOut - the path to store index
word2ID - the <word, index> pairs
ID2Word - the <index, word> pairs
tag2ID - the <tag, index> pairs
ID2Tag - the <index, tag> pairs
*/
void BatchManager::CreateIndex(const char *fileIn, const char *fileOut)
{
    CALL;


    wifstream in(fileIn, ios::in);
    CheckNTErrors(in.is_open(), "Failed to open the file!");
    in.imbue(utf8_locale);
    wstring line;

    // system("chcp 65001");

    /* the tag id starts from 0 */
    int tagId(0);
    while(getline(in, line)){
        if(!line.empty()){
            wstringstream ss(line);
            wstring word, tag;
            ss >> word >> tag;
            CheckNTErrors(!tag.empty(), "Failed to read tags from the training file!");

            /* allocate a DBG_NEW id for the tag */
            if(tag2Id.find(tag) == tag2Id.end()){
                tag2Id[tag] = tagId;
                id2Tag[tagId] = tag;
                tagId++;
            }
        }
    }
    in.close();

    /* store mappings */
    wofstream out(fileOut, ios::out);
    out.imbue(utf8_locale);
    for(const auto &map : id2Tag){
        out << map.first << "\t" << map.second << "\n";
    }
    out.close();


    OK;
}

/*
restore mappings from file
>>> file - the path of mappings
<<< word2ID - the <word, index> pairs
<<< ID2Word - the <index, word> pairs
<<< tag2ID - the <tag, index> pairs
<<< ID2Tag - the <index, tag> pairs
*/
void BatchManager::LoadIndex(const char *file)
{
    CALL;


    wifstream in(file, ios::in);
    in.imbue(utf8_locale);
    wstring line;
    while(getline(in, line)){
        int id;
        wstring tag;
        wstringstream ss(line);
        ss >> id >> tag;
        id2Tag[id] = tag;
        tag2Id[tag] = id;
    }
    in.close();
    LOG("mappings loaded");

    OK;
}

/*
load data to batches from a file
load words and tags if isTrain is true
only load words else
*/
BatchManager::BatchManager(Config *config, const char *mode)
{
    CALL;


    wifstream f;
    char errorMsg[1024] = {0};
    if(!(strcmp(mode, "train"))){
        f.open(config->trainFile.c_str(), ios::in);
        sprintf(errorMsg, "failed to open: %s", config->trainFile.c_str());
    }
    else if(!(strcmp(mode, "dev"))){
        f.open(config->devFile.c_str(), ios::in);
        sprintf(errorMsg, "failed to open: %s", config->devFile.c_str());
    }
    else if(!(strcmp(mode, "test"))){
        f.open(config->testFile.c_str(), ios::in);
        sprintf(errorMsg, "failed to open: %s", config->testFile.c_str());
    }
    else{
        CheckNTErrors(-1, "invalid mode!");
    }
    CheckNTErrors(f.is_open(), errorMsg);
    f.imbue(utf8_locale);

    /* prepare index for IDs and Words if there are no index */
    if(config->isTrain){
        CreateIndex(config->trainFile.c_str(), config->mapFile.c_str());
    }
    else{
        /* restore mappings from disk */
        LoadIndex(config->mapFile.c_str());
    }

    wstring line;
    vector<wstring> sentence;
    vector<vector<wstring>> batchSentences;
    vector<int> tags;

    /* the largest index is used for the fake tag 'Start' */
    tags.push_back(config->tagNum);

    vector<vector<int>> batchTags;
    while(getline(f, line)){

        /* collect words */
        if(!line.empty()){

            wstringstream ss(line);
            wstring word, tag;

            if(config->isSupervised){
                ss >> word >> tag;
                CheckNTErrors((!word.empty() && !tag.empty()), "invalid word or tag length!");
                tags.push_back(tag2Id[tag]);
            }
            else{
                ss >> word;
                CheckNTErrors((!word.empty()), "invalid word length!");
            }
            sentence.push_back(word);
        }

        /* collect sentence */
        else{
            batchSentences.push_back(sentence);
            sentence.clear();
            batchTags.push_back(tags);
            tags.clear();
            if(config->isSupervised)
                tags.push_back(config->tagNum);

            /* collect a mini-batch sentences */
            if(int(batchSentences.size()) == config->batchSize){
                allSentences.push_back(batchSentences);
                batchSentences.clear();
                if(config->isSupervised)
                    allTags.push_back(batchTags);
                batchTags.clear();
            }
        }
    }

    /* collect remained sentences */
    if(!batchSentences.empty()){
        allSentences.push_back(batchSentences);
        if(config->isSupervised)
            allTags.push_back(batchTags);
    }

    /* record the real length of sentences */
    for(const auto &ss : allSentences){
        vector<int> lengths;
        lengths.reserve(ss.size());
        for(const auto &s : ss){
            lengths.push_back(int(s.size()) + 1);
        }
        realLengths.push_back(lengths);
    }
    SortAndPad();
    f.close();


    OK;
}

/*
load parameters from the config file
the format is: <Param Value> (be similar to the yaml style)
notice that we use absolute paths
*/
Config::Config(const char *configFile)
{
    CALL;

    auto configString = ReadFile(configFile);

    useMem = configString["useMem"] != "false";
    isTrain = configString["isTrain"] != "false";
    isClean = configString["isClean"] != "false";
    useAdam = configString["useAdam"] != "false";
    useClip = configString["useClip"] != "false";
    isShuffled = configString["isShuffled"] != "false";
    isSupervised = configString["isSupervised"] != "false";
    useInitState = configString["useInitState"] != "false";

    devID = stoi(configString["devID"]);
    tagNum = stoi(configString["tagNum"]);
    charDim = stoi(configString["charDim"]);
    lstmUnit = stoi(configString["lstmUnit"]);
    maxEpoch = stoi(configString["maxEpoch"]);
    maxLength = stoi(configString["maxLength"]);
    batchSize = stoi(configString["batchSize"]);
    stepsCheck = stoi(configString["stepsCheck"]);

    dropoutP = stof(configString["dropoutP"]);
    clipLower = stof(configString["clipLower"]);
    clipUpper = stof(configString["clipUpper"]);
    adamBeta1 = stof(configString["adamBeta1"]);
    adamBeta2 = stof(configString["adamBeta2"]);
    adamDelta = stof(configString["adamDelta"]);
    adamBeta1T = stof(configString["adamBeta1T"]);
    adamBeta2T = stof(configString["adamBeta2T"]);
    learningRate = stof(configString["learningRate"]);

    logFile = configString["logFile"];
    mapFile = configString["mapFile"];
    embFile = configString["embFile"];
    ckptPath = configString["ckptPath"];
    devFile = configString["devFile"];
    testFile = configString["testFile"];
    trainFile = configString["trainFile"];
    preScript = configString["preScript"];
    tagScheme = configString["tagScheme"];
    evalScript = configString["evalScript"];
    recordFile = configString["recordFile"];
    ReadRecord();
    // CheckConfig();


    OK;
}

map<string, string> Config::ReadFile(const char* configFile)
{
    ifstream cfg(configFile, ios::in);
    CheckNTErrors(cfg.is_open(), "Failed to open the config/record file!");
    map<string, string> configString;
    string line;
    const string delimiterA = ":";
    const string delimiterB = "\t";
    while(getline(cfg, line)){
        if(line.size() > 4 && line.find("***") == string::npos){
            string param = line.substr(0, line.find_first_of(delimiterA));
            const string value = line.find('\n') != string::npos ?
                line.substr(line.find_last_of(delimiterB) + 1, line.size() - 1) :
                line.substr(line.find_last_of(delimiterB) + 1, line.size());
            configString[param] = value;
        }
    }
    cfg.close();
    return configString;
}

void Config::ReadRecord()
{
    if(!IsFileExists(recordFile.c_str()))
        MakeRecord();
    auto recordString = ReadFile(recordFile.c_str());
    record.loss = stof(recordString["loss"]);
    record.f1[0] = stof(recordString["trainF1"]);
    record.f1[1] = stof(recordString["devF1"]);
    record.f1[2] = stof(recordString["testF1"]);
    record.recall[0] = stof(recordString["trainRecall"]);
    record.recall[1] = stof(recordString["devRecall"]);
    record.recall[2] = stof(recordString["testRecall"]);
    record.accuracy[0] = stof(recordString["trainAccuracy"]);
    record.accuracy[1] = stof(recordString["devAccuracy"]);
    record.accuracy[2] = stof(recordString["testAccuracy"]);
    record.precision[0] = stof(recordString["trainPrecision"]);
    record.precision[1] = stof(recordString["devPrecision"]);
    record.precision[2] = stof(recordString["testPrecision"]);
}

void Config::MakeRecord()
{

    ofstream recordF(recordFile, ios::out);
    CheckNTErrors(recordF.is_open(), "Failed to open the record file!");

    recordF << "***Record***\n";

    recordF << "loss:" << "\t\t" << record.loss << "\n\n";

    recordF << "trainF1:" << "\t\t" << record.f1[0] << "\n";
    recordF << "trainRecall:" << "\t\t" << record.recall[0] << "\n";
    recordF << "trainAccuracy:" << "\t\t" << record.accuracy[0] << "\n";
    recordF << "trainPrecision:" << "\t\t" << record.precision[0] << "\n\n";

    recordF << "devF1:" << "\t\t" << record.f1[1] << "\n";
    recordF << "devRecall:" << "\t\t" << record.recall[1] << "\n";
    recordF << "devAccuracy:" << "\t\t" << record.accuracy[1] << "\n";
    recordF << "devPrecision:" << "\t\t" << record.precision[1] << "\n\n";

    recordF << "testF1:" << "\t\t" << record.f1[2] << "\n";
    recordF << "testRecall:" << "\t\t" << record.recall[2] << "\n";
    recordF << "testAccuracy:" << "\t\t" << record.accuracy[2] << "\n";
    recordF << "testPrecision:" << "\t\t" << record.precision[2];


    recordF.close();
}

/*
check and correct configurations after loading
*/
void Config::CheckConfig()
{
    CALL;

    string command = "python " + preScript;
    string msg = Exec(command.c_str());

    /* batchSize should be ranged from [1, sentenceNumber] */
    int validBatchSize = stoi(msg.substr(msg.find_last_of("batchSize:")));
    batchSize = validBatchSize >= batchSize ? batchSize : validBatchSize;

    OK;
}

/* execute system command and get the result */
std::string Exec(const char *cmd)
{
    CALL;


    std::array<char, 128> buffer{};
    std::string result;
#ifdef WIN32
    std::shared_ptr<FILE> pipe(_popen(cmd, "r"), _pclose);
#else
    std::shared_ptr<FILE> pipe(popen(cmd, "r"), pclose);
#endif // WIN32
    if(!pipe)
        throw std::runtime_error("popen() failed!");
    while(!feof(pipe.get())){
        if(fgets(buffer.data(), 128, pipe.get()) != nullptr)
            result += buffer.data();
    }


    OK;
    return result;
}

/*
dump ner results to a file
>>> paths - the decoded sequences
>>> data - the whole input data
>>> batch - the index of current data
>>> result - the file to store results
>>> isTraining - write real tags if it is TRUE
*/
void DumpResult(size_t batch, bool isTraining,
                const char *result, BatchManager &data,
                const vector<vector<int>> &paths)
{
    CALL;

    wofstream fout(result, ios::app);
    fout.imbue(utf8_locale);

    /* dump results to file */
    for(size_t i = 0; i < paths.size(); ++i){
        for(size_t j = 1; j < paths.at(i).size(); ++j){

            /* token */
            fout << data.allSentences[batch].at(i)[j - 1] << L"\t";

            /* correct tag */
            if(isTraining){
                int correctId = data.allTags[batch].at(i).at(j);
                fout << data.id2Tag[correctId] << L"\t";
            }
            int predictId = paths.at(i).at(j);

            /* predicted result */
            fout << data.id2Tag[predictId] << L"\n";
        }
        fout << L"\n";
    }
    fout << L"\n";
    fout.close();

    OK;
}

/*
create embeddings for one step in a mini-batch
>>> input - a tensor to store the input sentences
>>> sentences - sentences and tags in a mini-batch
>>> stepNum - the current step
>>> embDim - the size of embedding
>>> w2v - the word2vec struct
*/
void CreateInput(int embDim, int stepNum,
                 XTensor * input, Word2Vec& w2v,
                 const vector<vector<wstring>> &sentences)
{
    CALL;

    vector<wstring> batchStep;
    batchStep.reserve(sentences.size());
    for(const auto &sentence : sentences)
        batchStep.push_back(sentence[stepNum]);
    w2v.EmbeddingLookup(input, batchStep);

    OK;
}

} // namespace util