/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northeastern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com)
 */

#ifndef DATA_UTILITY_H_
#define DATA_UTILITY_H_

#include <array>
#include <codecvt>
#include <fstream>
#include <locale>
#include <map>
#include <stdexcept>
#include <vector>
#include "../../tensor/XGlobal.h"
#include "../../tensor/XTensor.h"

using namespace nts;
using namespace std;

namespace util
{
const locale utf8_locale = locale(locale(), new codecvt_utf8<wchar_t>());

enum class Score
{
    Train = 0,
    Dev = 1,
    Test = 2
};

/* records for SLTK */
struct Record
{
    /* record of lowest loss */
    DTYPE loss;

    /* records of best score on different data sets */
    DTYPE accuracy[3];
    DTYPE precision[3];
    DTYPE recall[3];
    DTYPE f1[3];
};

/* configurations for SLTK */
struct Config
{
    /* whether use mem pool */
    bool useMem;

    /* whether train the model */
    bool isTrain;

    /* whether clean the model */
    bool isClean;

    /* use adam or not */
    bool useAdam;

    /* use gradient clipping or not */
    bool useClip;

    /* whether shuffle the training data */
    bool isShuffled;

    /* whether initialize LSTM  states */
    bool useInitState;

    /* whether we know the correct tags */
    bool isSupervised;

    /* hyper parameters of adam*/
    DTYPE adamBeta1;
    DTYPE adamBeta2;
    DTYPE adamDelta;
    DTYPE adamBeta1T;
    DTYPE adamBeta2T;

    /* device id, -1 for CPU */
    int devID;

    /* number of tags */
    int tagNum;

    /* embedding size for characters */
    int charDim;

    /* length of cache, which should not be less than the max sentence length */
    int maxLength;

    /* number of hidden units in LSTM */
    int lstmUnit;

    /* maximum training epochs */
    int maxEpoch;

    /* batch size */
    int batchSize;

    /* steps per checkpoint */
    int stepsCheck;

    /* dropout probability */
    DTYPE dropoutP;

    /* lower bound for clipping */
    DTYPE clipLower;

    /* upper bound for clipping */
    DTYPE clipUpper;

    /* IOB/IOBES format */
    string tagScheme;

    /* learning rate */
    DTYPE learningRate;

    /* file for logging */
    string logFile;

    /* file for mappings */
    string mapFile;

    /* path to the pre-trained embedding */
    string embFile;

    /* path to save the model */
    string ckptPath;

    /* file for validation */
    string devFile;

    /* file for test */
    string testFile;

    /* file for training */
    string trainFile;

    /* pre-process script */
    string preScript;

    /* evaluation script */
    string evalScript;

    /* file for recording */
    string recordFile;

    /* records */
    Record record;

    explicit Config(const char *configFile);

    map<string, string> ReadFile(const char *configFile);
    void ReadRecord();
    void MakeRecord();
    void CheckConfig();
};

struct Word2Vec
{
    XTensor vec;
    map<wstring, int> word2Id;
    map<int, wstring> id2Word;

    Word2Vec(const char *word2VecFile, int embSize);
    void EmbeddingLookup(XTensor * embeddings, const vector<wstring> &sentence);
};

/* load data to batches */
class BatchManager
{
public:
    bool isTrain;
    map<int, wstring> id2Tag;                     // used for decoding
    map<wstring, int> tag2Id;                     // used for training
    vector<vector<int>> realLengths;              // real lengths of sentences
    vector<vector<vector<int>>> allTags;          // real tag sequences
    vector<vector<vector<wstring>>> allSentences; // all sentences

    void SortAndPad();
    void CreateIndex(const char *fileIn, const char *fileOut);
    void LoadIndex(const char *file);
    BatchManager(Config *config, const char *mode);
};

/* execute system command */
std::string Exec(const char *cmd);

/* check if a file exists */
bool IsFileExists(const char *fileName);

/* get the size of a file */
std::ifstream::pos_type FileSize(const char *filename);

/* dump ner results to a file */
void DumpResult(size_t batch, bool isTraining,
                const char *result, BatchManager &data,
                const vector<vector<int>> &paths);

/* load embeddings for sentences */
void CreateInput(int embDim, int stepNum,
                 XTensor * input, Word2Vec& w2v,
                 const vector<vector<wstring>> &sentences);

} // namespace util

#endif // DATA_UTILITY_H_
