/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northeastern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com)
 */

#include "SLTKUtility.h"
#include "SLTKLSTMCell.h"
#include "../../tensor/core/CHeader.h"
#include "../../tensor/function/FHeader.h"

using namespace ner;
using namespace util;
namespace ner
{
enum class Initializer
{
    zerosInitializer,
    onesInitializer,
    xavierInitializer
};

/*
constructor of LSTM
>>> myMem - memory pool
>>> myBatchSize - batch size
>>> myEmbSize- embedding dimension
>>> myLSTMUnit - hidden unit number
>>> myDevID - device id, use -1 for CPU
>>> maxTime - the maximum sentence length in data
>>> myIsTraining - indicates whether train the model
>>> myInitialState - indicates whether initialize states c and h
>>> MyIsGrad - indicates whether the object is used for storing gradients
*/
BasicLSTMCell::BasicLSTMCell(const int myDevID, XMem* myMem, const bool myIsGrad,
                             const bool myIsTraining, const bool myInitialState,
                             const int myEmbSize, const int maxTime,
                             const int myBatchSize, const int myLSTMUnit)
{
    CALL;
    mem = myMem;
    devID = myDevID;
    isGrad = myIsGrad;
    embSize = myEmbSize;
    lstmUnit = myLSTMUnit;
    batchSize = myBatchSize;
    isTraining = myIsTraining;
    useInitState = myInitialState;

    InitParams();
    cache = nullptr;
    buffer = nullptr;
    if(!isGrad)
        AllocateBuffer(maxTime);
    OK;
}

void BasicLSTMCell::AllocateBuffer(const int maxTime)
{
    CALL;

    /* cache layout: [i_t | tanh(input*w_xc + h_prev*w_hc + b_c) | c_t | o_t | h_t | input | tmp]  */
    cache = DBG_NEW XList(maxTime);

    int tmpDimSize[] = {int(batchSize), int(lstmUnit)};

    buffer = new XList(10);
    for(int i = 0; i < 10; ++i){
        XTensor *b = NewTensorBuf(2, tmpDimSize, X_FLOAT, 1.0F, devID, mem);
        buffer->Add(b);
    }

    for(int i = 0; i < maxTime; ++i){
        XList *stepCache = DBG_NEW XList(7);
        cache->Add(stepCache);

        for(int j = 0; j < 5; ++j){
            XTensor *tmp = NewTensorBuf(2, tmpDimSize, X_FLOAT, 1.0F, devID, mem);
            tmp->SetZeroAll();
            stepCache->Add(tmp);

            if(isTraining){
                if(j != 1){
                    XTensor *tmpGrad = NewTensorBuf(2, tmpDimSize, X_FLOAT, 1.0F, devID, mem);
                    tmp->grad = tmpGrad;
                }
            }
        }
        int inputDimSize[] = {int(batchSize), int(embSize)};
        XTensor *input = NewTensorBuf(2, inputDimSize, X_FLOAT, 1, devID, mem);
        stepCache->Add(input);
        XTensor *tmp = NewTensorBuf(2, tmpDimSize, X_FLOAT, 1.0F, devID, mem);
        tmp->SetZeroAll();
        stepCache->Add(tmp);
    }
    OK;
}

/* de-constructor */
BasicLSTMCell::~BasicLSTMCell()
{
    CALL;
    if(buffer){
        for(int i = 0; i < buffer->count; ++i){
            XTensor *tmp = static_cast<XTensor*>(buffer->Get(i));
            DelTensorBuf(tmp);
        }
    }
    if(cache){
        for(int i = 0; i < cache->count; ++i){
            XList *step = static_cast<XList*>(cache->Get(i));
            for(int j = 0; j < step->count; ++j){
                XTensor *tmp = static_cast<XTensor*>(step->Get(j));
                DelTensorBuf(tmp);
            }
            delete step;
        }
    }
    if(!isGrad){
        DelTensor(states[0]);
        DelTensor(states[1]);
    }
    delete buffer;
    delete cache;
    OK;
}

/*
init all tensors in a BasicLSTMCell
let x be the input whose shape is [batchSize, embDim]
then the shape of each parameter is:
w_x* ==> [embDim, lstmUnitsNum]
w_h* ==> [lstmUnitsNum, lstmUnitsNum]
w_c* ==> [lstmUnitsNum, lstmUnitsNum]
b_*  ==> [lstmUnitsNum]
*/
void BasicLSTMCell::InitParams()
{
    CALL;


    XTensor *params[] = {
        &w_xi, &w_xo, &w_xc,
        &w_hi, &w_ho, &w_hc,
        &w_ci, &w_co,
        &b_i, &b_c, &b_o
    };
    for(int i = 0; i < 3; ++i)
        InitTensor2D(params[i], embSize, lstmUnit, X_FLOAT, devID, mem);
    for(int i = 3; i < 8; ++i)
        InitTensor2D(params[i], lstmUnit, lstmUnit, X_FLOAT, devID, mem);
    for(int i = 8; i < 11; ++i)
        InitTensor2D(params[i], 1, lstmUnit, X_FLOAT, devID, mem);

    if(!isGrad){
        states[0] = NewTensor2D(batchSize, lstmUnit, X_FLOAT, devID, mem);
        states[1] = NewTensor2D(batchSize, lstmUnit, X_FLOAT, devID, mem);

        if(useInitState){
            LOG("use random initialized lstm states <c, h> ");
            _SetDataFanInOut(states[0]);
            _SetDataFanInOut(states[1]);
        }
    }

    /* by default, use xavier's  method to initialize parameters */
    if(!isGrad){
        for(auto& p : params){
            _SetDataFanInOut(p);
        }
        b_i.SetZeroAll();
        b_c.SetZeroAll();
        b_o.SetZeroAll();
        if(isTraining){
            /* allocate memory for gradient of parameters */
            for(int i = 0; i < 8; ++i){
                XTensor *grad = NewTensorBuf(params[i], devID, mem);
                grad->SetZeroAll();
                params[i]->grad = grad;
            }

            /* we will apply reduce_sum to the gradient of bias when updating */
            for(int i = 8; i < 11; ++i){
                XTensor *grad = NewTensor2D(batchSize, lstmUnit, X_FLOAT, devID, mem);
                grad->SetZeroAll();
                params[i]->grad = grad;
            }
        }
    }
    else{
        for(auto& p : params){
            p->SetZeroAll();
        }
    }

    OK;
}

/*
run one time step of lstm forward propagation
here is the implementation of 'coupled input and forget gates'
>>> cur - the current time step(started from 1)
>>> inputs - a [batch, embSize] tensor for the input
>>> state - the previous states of c and h
shapes of all tensors:
c_prev, h_prev, i_t, c_t, o_t, h_t  ==> [batch, lstmUnitNum]
w_xi, w_xo, w_xc                    ==> [embSize, lstmUnitNum]
w_hi, w_ci, w_ho, w_co, w_hc        ==> [lstmUnitNum, lstmUnitNum]
b_i, b_c, b_o                       ==> [lstmUnitNum, 1]
*/
void BasicLSTMCell::RunOneForwardStep(int timeStep)
{
    CALL;
    XList *stepCache = static_cast<XList *>(cache->Get(timeStep));

    XTensor *cPrev = states[0];
    XTensor *hPrev = states[1];

    /* the pureTmp is only used in this function */
    XTensor *iT = static_cast<XTensor*>(stepCache->Get(0));
    XTensor *tanhTmp = static_cast<XTensor*>(stepCache->Get(1));
    XTensor *cT = static_cast<XTensor*>(stepCache->Get(2));
    XTensor *oT = static_cast<XTensor*>(stepCache->Get(3));
    XTensor *hT = static_cast<XTensor*>(stepCache->Get(4));
    XTensor *xT = static_cast<XTensor*>(stepCache->Get(5));
    XTensor *pureTmp = static_cast<XTensor*>(stepCache->Get(6));
    XTensor *reuseO = oT;
    XTensor *reuseH = hT;
    for(auto &p : {iT,cT,oT,hT,pureTmp,tanhTmp})
        p->SetZeroAll();

    /* '@' is element-wise product and '*' is matrix multiplication */
    /* i_t = sigmoid(inputs * w_xi + h_prev * w_hi + c_prev * w_ci + b_i) */
    _MatrixMul2D(xT, X_NOTRANS, &w_xi, X_NOTRANS, iT);
    _MatrixMul2D(hPrev, X_NOTRANS, &w_hi, X_NOTRANS, reuseO);
    _MatrixMul2D(cPrev, X_NOTRANS, &w_ci, X_NOTRANS, reuseH);
    _SumMe(iT, reuseO);
    _SumMe(iT, reuseH);
    _SumDim(iT, &b_i, 1);
    _Sigmoid(iT, iT);

    /* c_t = (1 - i_t) @ c_prev + i_t @ tanh(inputs * w_xc + h_prev * w_hc + b_c) */
    _MatrixMul2D(xT, X_NOTRANS, &w_xc, X_NOTRANS, reuseO);
    _MatrixMul2D(hPrev, X_NOTRANS, &w_hc, X_NOTRANS, reuseH);
    _SumMe(tanhTmp, reuseO);
    _SumMe(tanhTmp, reuseH);
    _SumDim(tanhTmp, &b_c, 1);
    _HardTanH(tanhTmp, tanhTmp);
    _Multiply(tanhTmp, iT, hT);
    _ScaleAndShift(iT, cT, -1.0F, 1.0F);
    _MultiplyMe(cT, cPrev);
    _SumMe(cT, hT);

    /* o_t = sigmoid(inputs * w_xo + h_prev * w_ho + c_t * w_co + b_o) */
    _MatrixMul2D(xT, X_NOTRANS, &w_xo, X_NOTRANS, oT);
    _MatrixMul2D(hPrev, X_NOTRANS, &w_ho, X_NOTRANS, reuseH);
    _MatrixMul2D(cT, X_NOTRANS, &w_co, X_NOTRANS, pureTmp);
    _SumMe(oT, reuseH);
    _SumMe(oT, pureTmp);
    _SumDim(oT, &b_o, 1);
    _Sigmoid(oT, oT);

    /* h_t = o_t @ tanh(c_t) */
    _HardTanH(cT, hT);
    _MultiplyMe(hT, oT);

    /* update states of c and h */
    vector<int> srcIndex;
    srcIndex.reserve(cT->GetDim(0));
    for(int i = 0; i < cT->GetDim(0); ++i){
        srcIndex.push_back(i);
    }
    _CopyIndexed(cT, cPrev, 0, &srcIndex[0], cT->GetDim(0), &srcIndex[0], 1);
    _CopyIndexed(hT, hPrev, 0, &srcIndex[0], cT->GetDim(0), &srcIndex[0], 1);
    OK;
}

/*
merge h_t of all time step into outputs
>>> maxTimeBatch - the max time step in a mini-batch
>>> isReversed - indicates whether reverse the output
*/
void BasicLSTMCell::MergeOutput(int maxTimeBatch, bool isReversed)
{
    CALL;

    ExtendShape(maxTimeBatch + 1);
    XList steps(maxTimeBatch);

    /* notice that the first position is used for the initial states */
    if(!isReversed){
        for(int i = 1; i <= maxTimeBatch; ++i){
            XTensor *currentStepRes = static_cast<XTensor*>(static_cast<XList*>(cache->Get(i))->Get(4));
            steps.Add(currentStepRes);
        }
    }
    else{
        for(int i = maxTimeBatch; i > 0; --i){
            XTensor *currentStepRes = static_cast<XTensor*>(static_cast<XList*>(cache->Get(i))->Get(4));
            steps.Add(currentStepRes);
        }
    }

    InitTensor3D(&outputs, batchSize, maxTimeBatch, lstmUnit, X_FLOAT, devID, mem);
    if(isTraining && outputs.grad == nullptr)
        outputs.grad = NewTensorBuf(&outputs, devID, mem);
    if(isTraining)
        InitTensor3D(outputs.grad, batchSize, maxTimeBatch, lstmUnit, X_FLOAT, devID, mem);
    outputs.grad->SetZeroAll();
    _Merge(&steps, &outputs, 1);
    ShrinkShape(maxTimeBatch + 1);

    OK;
}

/* extend lstm outputs shape */
void BasicLSTMCell::ExtendShape(int maxTimeBatch) const
{


    for(int i = 0; i < maxTimeBatch; ++i){
        XTensor *currentStepRes = static_cast<XTensor*>(static_cast<XList*>(cache->Get(i))->Get(4));
        CheckNTErrors(currentStepRes->order == 2, "The order of result must be 2!");
        int shape[] = {currentStepRes->GetDim(0),1,currentStepRes->GetDim(1)};
        currentStepRes->Reshape(3, shape);
    }


    OK;
}

/* recover lstm outputs shape */
void BasicLSTMCell::ShrinkShape(int maxTimeBatch) const
{
    for(int i = 0; i < maxTimeBatch; ++i){
        XTensor *currentStepRes = static_cast<XTensor*>(static_cast<XList*>(cache->Get(i))->Get(4));
        CheckNTErrors(currentStepRes->order == 3, "The order of result must be 3!");
        int shape[] = {currentStepRes->GetDim(0),currentStepRes->GetDim(2)};
        currentStepRes->Reshape(2, shape);
    }
    OK;
}

/*
run one time step of lstm backward propagation
>>> timeStep - the current time step
>>> end - the final time step
>>> gradIN - gradient of h_t from the projection layer
>>> gradAccumulated - place to accumulate gradient of each time step
*/
void BasicLSTMCell::RunOneBackwardStep(int end, int timeStep, XTensor *gradIN, BasicLSTMCell *gradAccumulated)
{
    CALL;


    /*
    one step of backward propagation:
    dL/do_t = dL/dh_t @ tanh(c_t)
    dL/di_t = dL/dc_t @ (-c_prev + tanh(x_t * w_xc + h_prev * w_hc + b_c)

    if t==T, which is the final time step:
    dL/dh_t = df/dh_t
    dL/dc_t = df/dh_t @ [o_t @ (1-tanh^2(c_t)) + tanh(c_t) @ do_t/dc_t]

    else:
    dL/dh_t = df/dh_t + df/di_prev @ di_prev/dh_t + df/dc_prev @ dc_prev/dh_t + df/do_prev @ do_prev/dh_t
    dL/dc_t = df/dh_t @ [o_t @ (1-tanh^2(c_t)) + tanh(c_t) @ do_t/d_c_t] +
              df/dc_prev @ (1 - i_prev) + df/di_prev @ (di_prev/dc_t)
    */

    /* cache of the current time step  */
    XList *stepCache = static_cast<XList *>(cache->Get(timeStep));
    XTensor *xT = static_cast<XTensor*>(stepCache->Get(5));
    XTensor *iT = static_cast<XTensor*>(stepCache->Get(0));
    XTensor *cT = static_cast<XTensor*>(stepCache->Get(2));
    XTensor *oT = static_cast<XTensor*>(stepCache->Get(3));
    XTensor *hT = static_cast<XTensor*>(stepCache->Get(4));
    XTensor *tanhT = static_cast<XTensor*>(stepCache->Get(1));

    /* fetch buffer */
    for(int i = 0; i < buffer->count; ++i)
        static_cast<XTensor*>(buffer->Get(i))->SetZeroAll();

    XTensor *tanhC = static_cast<XTensor*>(buffer->Get(0));
    XTensor *dOdC = static_cast<XTensor*>(buffer->Get(1));
    XTensor * dSigmoid = static_cast<XTensor*>(buffer->Get(2));
    XTensor * temp = static_cast<XTensor*>(buffer->Get(3));
    XTensor * dCdI = static_cast<XTensor*>(buffer->Get(4));
    XTensor * dIdHprev = static_cast<XTensor*>(buffer->Get(5));
    XTensor * dIdCprev = static_cast<XTensor*>(buffer->Get(6));
    XTensor * dCdHprev = static_cast<XTensor*>(buffer->Get(7));
    XTensor * dCdCprev = static_cast<XTensor*>(buffer->Get(8));
    XTensor * dOdHprev = static_cast<XTensor*>(buffer->Get(9));

    /* cache of the previous time step  */
    XList * prevCache = static_cast<XList *>(cache->Get(timeStep - 1));
    XTensor *cPrev = static_cast<XTensor*>(prevCache->Get(2));
    XTensor *hPrev = static_cast<XTensor*>(prevCache->Get(4));

    /* dL/dh_t */
    if(timeStep != end)
        _SumMe(hT->grad, gradIN);
    else
        _CopyValues(gradIN, hT->grad);

    _HardTanH(cT, tanhC);

    /* dL/do_t */
    _Multiply(hT->grad, tanhC, oT->grad);

    /* dL/dc_t */
    _CopyValues(cT->grad, temp);
    _SigmoidBackward(oT, dSigmoid);
    _MultiplyMe(dSigmoid, tanhC);
    _MatrixMul(dSigmoid, X_NOTRANS, &w_co, X_TRANS, dOdC);
    _HardTanHBackward(tanhC, tanhC);
    _Multiply(oT, tanhC, cT->grad);
    _SumMe(cT->grad, dOdC);
    _MultiplyMe(cT->grad, hT->grad);
    if(timeStep != end)
        _SumMe(cT->grad, temp);

    /* dL/di_t */
    _Sub(tanhT, cPrev, dCdI);
    _Multiply(cT->grad, dCdI, iT->grad);

    /* gradient of [w_xi, w_ci, w_hi, b_i] */
    _SigmoidBackward(iT, b_i.grad);
    _MultiplyMe(b_i.grad, iT->grad);
    _MatrixMul(xT, X_TRANS, b_i.grad, X_NOTRANS, w_xi.grad);
    _MatrixMul(cPrev, X_TRANS, b_i.grad, X_NOTRANS, w_ci.grad);
    _MatrixMul(hPrev, X_TRANS, b_i.grad, X_NOTRANS, w_hi.grad);

    /* gradient of [w_xc, w_hc, b_c] */
    _HardTanHBackward(tanhT, b_c.grad);
    _MultiplyMe(b_c.grad, iT);
    _MultiplyMe(b_c.grad, cT->grad);
    _MatrixMul(xT, X_TRANS, b_c.grad, X_NOTRANS, w_xc.grad);
    _MatrixMul(hPrev, X_TRANS, b_c.grad, X_NOTRANS, w_hc.grad);

    /* gradient of [w_xo, w_co, w_ho, b_o] */
    _SigmoidBackward(oT, b_o.grad);
    _MultiplyMe(b_o.grad, oT->grad);
    _MatrixMul(xT, X_TRANS, b_o.grad, X_NOTRANS, w_xo.grad);
    _MatrixMul(cT, X_TRANS, b_o.grad, X_NOTRANS, w_co.grad);
    _MatrixMul(hPrev, X_TRANS, b_o.grad, X_NOTRANS, w_ho.grad);

    /* accumulate gradient of current step */
    AccumulateGrads(gradAccumulated);

    /* pass gradients to the next step
     * such as dh_t/d_h_t-1, put it into h_t-1.grad  */
    if(timeStep != 1){
        hPrev->grad->SetZeroAll();
        cPrev->grad->SetZeroAll();

        /* part i: dLi/dCprev, dLi/dHprev */
        _MatrixMul(b_i.grad, X_NOTRANS, &w_hi, X_TRANS, dIdHprev);
        _MatrixMul(b_i.grad, X_NOTRANS, &w_hc, X_TRANS, dIdCprev);

        /* part c: dLc/dCprev, dLc/dHprev */
        _MatrixMul(b_c.grad, X_NOTRANS, &w_hc, X_TRANS, dCdHprev);
        _ScaleAndShift(iT, dCdCprev, -1.0F, 1.0F);
        _MultiplyMe(dCdCprev, cT->grad);

        /* part o: dLo/dHprev */
        _MatrixMul(b_o.grad, X_NOTRANS, &w_ho, X_TRANS, dOdHprev);

        _SumMe(cPrev->grad, dCdCprev);
        _SumMe(cPrev->grad, dIdCprev);
        _SumMe(hPrev->grad, dOdHprev);
        _SumMe(hPrev->grad, dCdHprev);
        _SumMe(hPrev->grad, dIdHprev);
    }
    OK;
}

/* accumulate gradient for each param across time steps */
void BasicLSTMCell::AccumulateGrads(BasicLSTMCell *grad) const
{
    _SumMe(&grad->w_xi, w_xi.grad);
    _SumMe(&grad->w_xc, w_xc.grad);
    _SumMe(&grad->w_xo, w_xo.grad);
    _SumMe(&grad->w_ci, w_ci.grad);
    _SumMe(&grad->w_co, w_co.grad);
    _SumMe(&grad->w_hi, w_hi.grad);
    _SumMe(&grad->w_hc, w_hc.grad);
    _SumMe(&grad->w_ho, w_ho.grad);

    /* apply reduce_sum to the bias gradient */
    int biasDimSize[] = {grad->b_i.GetSize()};
    int gradDimSize[] = {1, grad->b_i.GetSize()};

    XTensor *buffer = NewTensorBuf(&(grad->b_i), devID, mem);
    XTensor *bias[]{b_i.grad, b_c.grad, b_o.grad};
    XTensor *biasGrads[]{&grad->b_i, &grad->b_c, &grad->b_o};

    for(int i = 0; i < 3; ++i){
        buffer->Reshape(1, biasDimSize);
        _ReduceSum(bias[i], buffer, 0);
        buffer->Reshape(2, gradDimSize);
        _SumMe(biasGrads[i], buffer);
    }
    DelTensorBuf(buffer);
    OK;
}

/* constructor */
BiLSTMLayer::BiLSTMLayer(XMem *myMem, Config *myConfig)
{
    CALL;

    config = myConfig;
    mem = myMem;
    devID = config->devID;

    forwardCell = DBG_NEW BasicLSTMCell(config->devID, mem, false, true, config->useInitState,
                                    config->charDim, config->maxLength, config->batchSize, config->lstmUnit);
    backwardCell = DBG_NEW BasicLSTMCell(config->devID, mem, false, true, config->useInitState,
                                     config->charDim, config->maxLength, config->batchSize, config->lstmUnit);
    forwardGrad = nullptr;
    backwardGrad = nullptr;

    if(config->isTrain){
        forwardGrad = DBG_NEW BasicLSTMCell(config->devID, mem, true, false, false,
                                        config->charDim, 0, config->batchSize, config->lstmUnit);
        backwardGrad = DBG_NEW BasicLSTMCell(config->devID, mem, true, false, false,
                                         config->charDim, 0, config->batchSize, config->lstmUnit);
    }
    OK;
}

/* de-constructor */
BiLSTMLayer::~BiLSTMLayer()
{
    CALL;
    delete forwardCell;
    delete backwardCell;
    delete forwardGrad;
    delete backwardGrad;
    OK;
}

/*
biLSTM forward propagation (in a mini-batch)
that will compute the outputs in two directions
>>> w2v - the embeddings lookup table
>>> predict - a [batchSize, maxTime, tagNum] tensor for score of tags
>>> sentences - a [batchSize, maxTime] matrix for sentences
*/
void BiLSTMLayer::Forward(Word2Vec &w2v, XTensor *biLSTMOutput, const vector<vector<wstring>> &sentences)
{
    CALL;
    const int maxTimeInBatch = sentences.front().size();
    BasicLSTMCell *lstmCell[] = {forwardCell, backwardCell};

    /* lstm forward propagation in two directions */
    for(int i = 0; i < 2; ++i){

        lstmCell[i]->states[0]->SetZeroAll();
        lstmCell[i]->states[1]->SetZeroAll();

        for(int forwardStep = 0; forwardStep < maxTimeInBatch; ++forwardStep){

            /* use i=0 for forward sequences and i=1 for backward sequences */
            /* the first position of cache is used for the initial states   */
            const int timeStep = (!i) ? forwardStep : (maxTimeInBatch - forwardStep - 1);

            XList *cache = static_cast<XList*>(lstmCell[i]->cache->Get(forwardStep + 1));
            XTensor *input = static_cast<XTensor*>(cache->Get(5));

            CreateInput(forwardCell->embSize, timeStep, input, w2v, sentences);

            _Dropout(input, input, clock(), config->dropoutP);
            lstmCell[i]->RunOneForwardStep(forwardStep + 1);
        }
    }

    forwardCell->MergeOutput(maxTimeInBatch, false);
    backwardCell->MergeOutput(maxTimeInBatch, true);

    /* concatenate forward and backward lstm outputs into one */
    XList lstmOutputs(2);
    lstmOutputs.Add(&forwardCell->outputs);
    lstmOutputs.Add(&backwardCell->outputs);

    InitTensor3D(biLSTMOutput, forwardCell->outputs.GetDim(0), maxTimeInBatch, forwardCell->lstmUnit * 2, X_FLOAT, devID, mem);
    _Concatenate(&lstmOutputs, biLSTMOutput, 2);
    OK;
}

/*
biLSTM backward propagation (in a mini-batch)
that will compute the gradient in two directions
*/
void BiLSTMLayer::Backward(XTensor* biLSTMOutput)
{
    CALL;
    BasicLSTMCell *lstmCell[] = {forwardCell,backwardCell};
    BasicLSTMCell *lstmGrad[] = {forwardGrad,backwardGrad};

    int stepShape[]{biLSTMOutput->GetDim(0), 1, biLSTMOutput->GetDim(2) / 2};
    XTensor *stepGrad = NewTensorBuf(3, stepShape, X_FLOAT, 1, devID, mem);
    stepGrad->SetZeroAll();

    /* bi-directional lstm back-propagation */
    for(int i = 0; i < 2; ++i){

        const int maxTimeInBatch = lstmCell[i]->outputs.GetDim(1);

        /* reuse of single LSTM outputs to store its gradient */
        const int low = i * (biLSTMOutput->GetDim(2) / 2);
        const int high = (i + 1) * (biLSTMOutput->GetDim(2) / 2);
        _SelectRange(biLSTMOutput, &lstmCell[i]->outputs, 2, low, high);
        int bigDim[]{biLSTMOutput->GetDim(0), 1, biLSTMOutput->GetDim(2) / 2};
        int smallDim[]{biLSTMOutput->GetDim(0), biLSTMOutput->GetDim(2) / 2};

        for(int j = 0; j < maxTimeInBatch; ++j){
            stepGrad->Reshape(3, bigDim);
            const int gradTimeStep = i ? j : maxTimeInBatch - j - 1;
            const int cacheTimeStep = maxTimeInBatch - j;
            _SelectRange(&lstmCell[i]->outputs, stepGrad, 1, gradTimeStep, gradTimeStep + 1);
            stepGrad->Reshape(2, smallDim);
            lstmCell[i]->RunOneBackwardStep(maxTimeInBatch, cacheTimeStep, stepGrad, lstmGrad[i]);
        }
    }

    DelTensorBuf(stepGrad);
    OK;
}
} // namespace ner