/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northeastern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com)
 */

#ifndef LSTM_CELL_H_
#define LSTM_CELL_H_


#include "SLTKDataUtility.h"

using namespace nts;
using namespace std;
using namespace util;
namespace ner
{

/* coupled forget and input gate lstm network */
struct BasicLSTMCell
{
    /* train the model or not */
    bool isTraining;

    /* use random states or not */
    bool useInitState;

    /* just store gradient if isGrad is true */
    bool isGrad;

    /* device id, -1 for CPU */
    int devID;

    /* memory pool */
    XMem * mem;

    /* cache of all time steps */
    XList *cache;

    /* buffer in backward-propagation */
    XList *buffer;

    /* collections of all h_t in a mini-batch */
    XTensor outputs;

    /* batch size */
    int batchSize;

    /* number of LSTM unit */
    int lstmUnit;

    /* input embedding dimension */
    int embSize;

    /* input gate weights */
    XTensor w_xi;
    XTensor w_hi;
    XTensor w_ci;

    /* output gate weights */
    XTensor w_xo;
    XTensor w_ho;
    XTensor w_co;

    /* cell weights  */
    XTensor w_xc;
    XTensor w_hc;

    /* bias vectors  */
    XTensor b_i;
    XTensor b_c;
    XTensor b_o;

    /* we treat the initial states as variables as well */
    XTensor *states[2];

    explicit BasicLSTMCell(int myDevID, XMem* myMem, bool myIsGrad,
                           bool myIsTraining, bool myInitialState,
                           int myEmbSize, int maxTime,
                           int myBatchSize, int myLSTMUnit);
    void AllocateBuffer(int maxTime);

    ~BasicLSTMCell();

    /* initialize all variables with random values */
    void InitParams();

    /* forward propagation through time */
    void RunOneForwardStep(int timeStep);

    /* merge h_t of all time step into a big tensor */
    void MergeOutput(int maxTimeBatch, bool isReversed);

    /* extend lstm outputs shape */
    void ExtendShape(int maxTimeBatch) const;

    /* recover lstm outputs shape */
    void ShrinkShape(int maxTimeBatch) const;

    /* backward propagation through time */
    void RunOneBackwardStep(int end, int timeStep, XTensor *gradIN, BasicLSTMCell *gradAccumulated);

    /* accumulate gradient for each param across time steps */
    void AccumulateGrads(BasicLSTMCell *grad) const;
};

struct BiLSTMLayer
{
    int devID;

    XMem *mem;

    Config *config;

    /* forward LSTM cell */
    BasicLSTMCell *forwardCell;

    /* forward LSTM cell gradient */
    BasicLSTMCell *forwardGrad;

    /* backward LSTM cell */
    BasicLSTMCell *backwardCell;

    /* backward LSTM cell gradient */
    BasicLSTMCell *backwardGrad;

    BiLSTMLayer(XMem *myMem, Config *myConfig);

    ~BiLSTMLayer();

    /* biLSTM forward propagation (in a mini-batch) */
    void Forward(Word2Vec &w2v,
                 XTensor *biLSTMOutput,
                 const vector<vector<wstring>> &sentences);

    /* biLSTM backward propagation (in a mini-batch) */
    void Backward(XTensor *biLSTMOutput);
};

} // namespace ner

#endif // LSTM_CELL_H_