/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northeastern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com)
 */

#include "SLTKModel.h"
#include "SLTKUtility.h"
#include "../../tensor/core/CHeader.h"

using namespace util;

namespace ner
{

/*
model forward propagation (in a mini-batch)
>>> w2v - the embeddings lookup table
>>> realLengths - a [batchSize] list, real sequence length
>>> tagIndices - a [batchSize, maxTime+1] list, correct tags
>>> sentences -  a [batchSize, maxTime+1] list, input sequences
<<< loss of a mini-batch
*/
DTYPE BiLSTMCRFModel::Forward(Word2Vec &w2v,
                              const vector<int> &realLengths,
                              const vector<vector<int>> &tagIndices,
                              const vector<vector<wstring>> &sentences)
{
    CALL;

    /* reset caches in all layers */
    ResetCache();

    XTensor *biLSTMOutput = &cache[0];
    XTensor *projectScores = &cache[1];
    XTensor *extendedScores = &cache[2];

    lstm->Forward(w2v, biLSTMOutput, sentences);

    projection->Forward(biLSTMOutput, projectScores);

    /* extend projected results to suit crf */
    util::_PaddingScores(projectScores, extendedScores);

    /*projectScores->Dump(stderr);
    extendedScores->Dump(stderr);*/

    OK;
    return -crf->LogLikelihood(extendedScores, tagIndices, realLengths);
}

/*
run one forward step of the model(mini-batch)
>>> tagIndices - a [batchSize, maxTime+1] table records correct tags
>>> realLengths - a [batchSize, realLength+1] table records sentence length
*/
void BiLSTMCRFModel::Backward(const vector<int> &realLengths, const vector<vector<int>> &tagIndices)
{
    CALL;


    XTensor *biLSTMOutput = &cache[0];
    XTensor *projectScores = &cache[1];
    XTensor *extendedScores = &cache[2];

    /* allocate memory for gradient of cache */
    if(config->isTrain){
        for(auto&c : cache){
            if(c.grad == nullptr)
                c.grad = NewTensorBuf(&c, config->devID, mem);
            c.grad->SetZeroAll();
        }
    }

    /* back-propagation in the crf layer */
    crf->Backward(extendedScores, realLengths, tagIndices);

    /* back-propagation in the projection layer */
    _ShrinkScores(extendedScores->grad, projectScores);
    int shape[]{projectScores->GetDim(0)*projectScores->GetDim(1),projectScores->GetDim(2)};
    projectScores->Reshape(2, shape);

    /* notice: reuse of 'biLSTMOutput' and 'projectScores' to store gradient */
    projection->Backward(biLSTMOutput, projectScores, biLSTMOutput);

    /* back-propagation in the bi-lstm layer */
    int shape2[]{extendedScores->GetDim(0), extendedScores->GetDim(1) - 1, biLSTMOutput->GetDim(1)};
    biLSTMOutput->Reshape(3, shape2);

    lstm->Backward(biLSTMOutput);

    OK;
}

/*
apply gradient to variables in a layer
>>> lr - the learning rate
*/
void BiLSTMCRFModel::Update(DTYPE lr)
{
    CALL;

    XTensor *checkList[]
    {
        crf->transition.grad,
        projection->outputW.grad, projection->outputB.grad,
        projection->hiddenW.grad, projection->hiddenB.grad
    };
    //for(auto &g : checkList){
    //    _ClipMe(g, -5.0F, 5.0F);
    //}
    BasicLSTMCell *lstmCell[] = {lstm->forwardCell,lstm->backwardCell};
    BasicLSTMCell *lstmGrad[] = {lstm->forwardGrad,lstm->backwardGrad};

    _SumMe(&crf->transition, crf->transition.grad, -lr);
    _SumMe(&projection->outputW, projection->outputW.grad, -lr);
    _SumMe(&projection->outputB, projection->outputB.grad, -lr);
    _SumMe(&projection->hiddenW, projection->hiddenW.grad, -lr);
    _SumMe(&projection->hiddenB, projection->hiddenB.grad, -lr);

    for(auto &g : checkList)
        g->SetZeroAll();

    for(int i = 0; i < 2; ++i){
        XTensor *grads[] = {
            &lstmGrad[i]->w_xi, &lstmGrad[i]->w_ci, &lstmGrad[i]->w_hi,
            &lstmGrad[i]->w_xc, &lstmGrad[i]->w_hc,
            &lstmGrad[i]->w_xo, &lstmGrad[i]->w_co, &lstmGrad[i]->w_ho,
            &lstmGrad[i]->b_i, &lstmGrad[i]->b_c, &lstmGrad[i]->b_o
        };
        //for(auto &g : grads){
        //    _ClipMe(g, -5.0F, 5.0F);
        //}
        _SumMe(&(lstmCell[i]->w_xi), &lstmGrad[i]->w_xi, -lr);
        _SumMe(&(lstmCell[i]->w_ci), &lstmGrad[i]->w_ci, -lr);
        _SumMe(&(lstmCell[i]->w_hi), &lstmGrad[i]->w_hi, -lr);
        _SumMe(&(lstmCell[i]->w_xc), &lstmGrad[i]->w_xc, -lr);
        _SumMe(&(lstmCell[i]->w_hc), &lstmGrad[i]->w_hc, -lr);
        _SumMe(&(lstmCell[i]->w_xo), &lstmGrad[i]->w_xo, -lr);
        _SumMe(&(lstmCell[i]->w_co), &lstmGrad[i]->w_co, -lr);
        _SumMe(&(lstmCell[i]->w_ho), &lstmGrad[i]->w_ho, -lr);
        _SumMe(&(lstmCell[i]->b_i), &lstmGrad[i]->b_i, -lr);
        _SumMe(&(lstmCell[i]->b_c), &lstmGrad[i]->b_c, -lr);
        _SumMe(&(lstmCell[i]->b_o), &lstmGrad[i]->b_o, -lr);

        for(auto &grad : grads){
            grad->SetZeroAll();
        }
    }
    ResetCache();


    OK;
}

/* clear cache of all layers */
void BiLSTMCRFModel::ResetCache()
{
    CALL;





    OK;
}

/*
evaluate the model
notice that it will run python scripts
>>> mode - 'train' for training, 'dev' for validation, 'test' for test
>>> w2v - word/char/subword embeddings
<<< score - a DTYPE value indicates score on data
input - the path of data used for evaluation
output - the path of results for evaluation
isTrain - whether save the model
*/
bool BiLSTMCRFModel::Evaluate(const char *mode, Word2Vec &w2v)
{
    CALL;

    string output;
    int modeVal(0);
    if(!strcmp(mode, "train")){
        output = config->trainFile;
        modeVal = 0;
    }
    else if(!strcmp(mode, "test")){
        output = config->testFile;
        modeVal = 2;
    }
    else if(!strcmp(mode, "dev")){
        output = config->devFile;
        modeVal = 1;
    }
    else
        CheckNTErrors(-1, "invalid mode!");
    output += ".output";

    BatchManager evalData(config, mode);

    XTensor *biLSTMOutput = &cache[0];
    XTensor *projectScores = &cache[1];
    XTensor *extendedScores = &cache[2];
    biLSTMOutput->SetZeroAll();
    projectScores->SetZeroAll();
    extendedScores->SetZeroAll();

    /* overwrite the output file */
    if(IsFileExists(output.c_str())){
        wofstream fout(output, ios::out);
        fout.close();
    }

    for(size_t batch = 0; batch < evalData.allSentences.size(); ++batch){

        /* reset caches in all layers */
        ResetCache();

        lstm->Forward(w2v, biLSTMOutput, evalData.allSentences.at(batch));

        projection->Forward(biLSTMOutput, projectScores);

        /* extend projected results to suit crf */
        _PaddingScores(projectScores, extendedScores);

        vector<vector<int>> tagSequences(evalData.allSentences.at(batch).size());

        crf->Decode(tagSequences, extendedScores, evalData.realLengths[batch]);

        DumpResult(batch, config->isSupervised, output.c_str(), evalData, tagSequences);
    }

    /* call python script to evaluate the results */
    string command("python ");
    command += config->evalScript;
    command += " ";
    command += output;
    string message = Exec(command.c_str());

    string line(message.substr(0, message.find_first_of('\n')));
    const size_t posA = line.find("accuracy:") + 11;
    const size_t posP = line.find_last_of("precision: ");
    const size_t posR = line.find_last_of("recall: ");
    const size_t posF = line.find("FB1:") + 5;
    const DTYPE scoreA = stof(line.substr(posA, 5));
    const DTYPE scoreP = stof(line.substr(posP, 5));
    const DTYPE scoreR = stof(line.substr(posR, 5));
    const DTYPE scoreF = stof(line.substr(posF, 5));

    /* notice that we independently record different best scores
     * that means different best scores can be obtained at different epoch */
    const auto Update = [](DTYPE &rawScore, const DTYPE &newScore)
    {
        if(newScore > rawScore){
            rawScore = newScore;
            return true;
        }
        return false;
    };

    Update(config->record.accuracy[modeVal], scoreA);
    Update(config->record.precision[modeVal], scoreP);
    Update(config->record.recall[modeVal], scoreR);

    OK;
    return Update(config->record.f1[modeVal], scoreF);
}

/*
constructor of the model
>>> cfg - the configurations of the whole model
*/
BiLSTMCRFModel::BiLSTMCRFModel(Config *cfg)
{
    CALL;


    config = cfg;

    /* allocate mem pool */
    mem = nullptr;
    //mem = DBG_NEW XMem(config->devID, FREE_ON_THE_FLY, static_cast<MTYPE>(MILLION) * 256, 1024, MILLION * 128);

    /* biLSTM layer */
    lstm = DBG_NEW BiLSTMLayer(mem, config);

    /* projection layer */
    projection = DBG_NEW ProjectLayer(config->lstmUnit, config->tagNum, config->devID, mem, config->isTrain);

    /* CRF layer */
    crf = DBG_NEW CRF(mem, config);

    if(IsFileExists(config->ckptPath.c_str()) && FileSize(config->ckptPath.c_str()) > 1000){

        /* if there is already a model file on the disk, then restore it */
        RestoreModel(config->ckptPath.c_str());

    }
    else{

        /* init parameters in the model */
        LOG("model file not found, initialize the model");

        StoreModel(config->ckptPath.c_str());
    }



    OK;
}

BiLSTMCRFModel::~BiLSTMCRFModel()
{
    delete mem;
    delete lstm;
    delete projection;
    delete crf;
}

/*
get parameter matrices
>> list - the list that keeps the parameter matrices
*/
void BiLSTMCRFModel::GetParams(XList& list) const
{


    list.Clear();

    list.Add(&crf->transition);

    list.Add(&projection->hiddenW);
    list.Add(&projection->hiddenB);
    list.Add(&projection->outputW);
    list.Add(&projection->outputB);

    BasicLSTMCell *lstmGrads[] = {lstm->forwardGrad, lstm->backwardGrad};
    for(auto &lstmGrad : lstmGrads){
        list.Add(&lstmGrad->w_xi);
        list.Add(&lstmGrad->w_xc);
        list.Add(&lstmGrad->w_xo);
        list.Add(&lstmGrad->w_ci);
        list.Add(&lstmGrad->w_co);
        list.Add(&lstmGrad->w_hi);
        list.Add(&lstmGrad->w_hc);
        list.Add(&lstmGrad->w_ho);
        list.Add(&lstmGrad->b_i);
        list.Add(&lstmGrad->b_c);
        list.Add(&lstmGrad->b_o);
    }


    OK;
}

/*
dump the model to the disk space
>> fn - where to keep the model
*/
void BiLSTMCRFModel::StoreModel(const char *fn) const
{


    FILE *ckpt = fopen(fn, "wb");
    CheckNTErrors(ckpt, "Cannot open the model file");

    BasicLSTMCell *lstmCell[] = {lstm->forwardCell,lstm->backwardCell};

    /* dump forward and backward cells in the lstm layer to file */
    for(auto &i : lstmCell){
        i->w_xi.Dump(ckpt, "lstm cell w_xi:");
        i->w_xo.Dump(ckpt, "lstm cell w_xo:");
        i->w_xc.Dump(ckpt, "lstm cell w_xc:");
        i->w_hi.Dump(ckpt, "lstm cell w_hi:");
        i->w_ho.Dump(ckpt, "lstm cell w_ho:");
        i->w_hc.Dump(ckpt, "lstm cell w_hc:");
        i->w_ci.Dump(ckpt, "lstm cell w_ci:");
        i->w_co.Dump(ckpt, "lstm cell w_co:");
        i->b_i.Dump(ckpt, "lstm cell b_i:");
        i->b_c.Dump(ckpt, "lstm cell b_c:");
        i->b_o.Dump(ckpt, "lstm cell b_o:");
    }

    /* dump weights in the projection layer to file */
    projection->hiddenW.Dump(ckpt, "projection hiddenW:");
    projection->hiddenB.Dump(ckpt, "projection hiddenB:");
    projection->outputW.Dump(ckpt, "projection scoresW:");
    projection->outputB.Dump(ckpt, "projection scoresB:");

    /* dump transitions in the crf layer to file */
    crf->transition.Dump(ckpt, "crf transition:");

    fclose(ckpt);
    LOG("ner model saved");


    OK;
}

/*
read the model from the disk space
>> fn - where to keep the model
*/
void BiLSTMCRFModel::RestoreModel(const char *fn) const
{


    FILE *ckpt = fopen(fn, "rb");
    CheckNTErrors(ckpt, "Cannot open the model file");

    BasicLSTMCell *lstmCell[] = {lstm->forwardCell,lstm->backwardCell};

    /* read forward and backward cells in the lstm layer from file */
    for(auto &i : lstmCell){
        i->w_xi.Read(ckpt, "lstm cell w_xi:");
        i->w_xo.Read(ckpt, "lstm cell w_xo:");
        i->w_xc.Read(ckpt, "lstm cell w_xc:");
        i->w_hi.Read(ckpt, "lstm cell w_hi:");
        i->w_ho.Read(ckpt, "lstm cell w_ho:");
        i->w_hc.Read(ckpt, "lstm cell w_hc:");
        i->w_ci.Read(ckpt, "lstm cell w_ci:");
        i->w_co.Read(ckpt, "lstm cell w_co:");
        i->b_i.Read(ckpt, "lstm cell b_i:");
        i->b_c.Read(ckpt, "lstm cell b_c:");
        i->b_o.Read(ckpt, "lstm cell b_o:");
    }

    /* read weights in the projection layer from file */
    projection->hiddenW.Read(ckpt, "projection hiddenW:");
    projection->hiddenB.Read(ckpt, "projection hiddenB:");
    projection->outputW.Read(ckpt, "projection scoresW:");
    projection->outputB.Read(ckpt, "projection scoresB:");

    /* read transitions in the crf layer from file */
    crf->transition.Read(ckpt, "crf transition:");

    fclose(ckpt);

    LOG("ner model loaded");


    OK;
}

} // namespace ner