/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northeastern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com)
 */

#include "SLTKUtility.h"
#include "SLTKProjection.h"
#include "../../tensor/core/CHeader.h"
#include "../../tensor/function/HardTanH.h"

using namespace util;

namespace ner
{

/* constructor */
ProjectLayer::ProjectLayer(const int inputDim, const int outputDim,
                           const int myDevID, XMem *myMem, const bool myIsTraining)
{
    CALL;

    mem = myMem;
    devID = myDevID;
    isTraining = myIsTraining;

    InitTensor2D(&hiddenW, inputDim * 2, inputDim, X_FLOAT, devID, mem);
    InitTensor2D(&hiddenB, inputDim, 1, X_FLOAT, devID, mem);
    InitTensor2D(&outputW, inputDim, outputDim, X_FLOAT, devID, mem);
    InitTensor2D(&outputB, outputDim, 1, X_FLOAT, devID, mem);
    _SetDataFanInOut(&hiddenW);
    _SetDataFanInOut(&outputW);
    hiddenB.SetZeroAll();
    outputB.SetZeroAll();

    if(isTraining){
        hiddenW.grad = NewTensorBuf(&hiddenW, devID, mem);
        hiddenB.grad = NewTensorBuf(&hiddenB, devID, mem);
        outputW.grad = NewTensorBuf(&outputW, devID, mem);
        outputB.grad = NewTensorBuf(&outputB, devID, mem);
    }

    OK;
}

/*
a projection between the lstm layer and the crf layer
>>> input - a [batchSize*maxTime, inputDim*2] tensor
<<< output - a [batchSize, maxTime, outputDim] tensor
*/
void ProjectLayer::Forward(XTensor *input, XTensor *output)
{
    CALL;

    const int batchSize = input->GetDim(0);
    XTensor *tmp = &cache[0];
    XTensor *hidden = &cache[1];

    /* hidden = tanh(input * hiddenW + hiddenB) */
    int shape[]{input->GetSize() / input->GetDim(2), input->GetDim(2)};
    input->Reshape(2, shape);

    InitTensor2D(tmp, input->GetDim(0), hiddenW.GetDim(1), X_FLOAT, devID, mem);
    InitTensor2D(hidden, input->GetDim(0), hiddenW.GetDim(1), X_FLOAT, devID, mem);
    tmp->SetZeroAll();
    hidden->SetZeroAll();

    _MatrixMul(input, X_NOTRANS, &hiddenW, X_NOTRANS, tmp);
    _SumDim(tmp, &hiddenB, 1);
    _HardTanH(tmp, hidden);

    /* output = hidden * outputW + outputB */
    InitTensor2D(output, hidden->GetDim(0), outputW.GetDim(1), X_FLOAT, devID, mem);
    _MatrixMul2D(hidden, X_NOTRANS, &outputW, X_NOTRANS, output);
    _SumDim(output, &outputB, 1);

    /* reshape the result */
    int shape2[]{batchSize, output->GetSize() / (batchSize * outputW.GetDim(1)), outputW.GetDim(1)};
    output->Reshape(3, shape2);

    OK;
}


/*
backward propagation in the projection layer
>>> input - a [batchSize*maxTime, outputDim*2] tensor indicates the input of this layer
>>> gradIN - a [maxTime, outputDim] tensor indicates gradient passed to this layer
>>> gradOUT - a [batchSize*maxTime, inputDim*2] tensor indicates gradient of the input
notice that 'input' and 'gradOUT' may be the same tensor to save memory
*/
void ProjectLayer::Backward(XTensor *input, XTensor *gradIn, XTensor *gradOut)
{
    CALL;

    /* allocate memory for gradient of cache */
    if(isTraining){
        for(auto&c : cache){
            if(c.grad == nullptr)
                c.grad = NewTensorBuf(&c, devID, mem);
            c.grad->SetZeroAll();
        }
    }
    /* gradient of hidden */
    XTensor *hidden = &cache[1];
    InitTensor(hidden->grad, hidden);

    _MatrixMul(gradIn, X_NOTRANS, &outputW, X_TRANS, hidden->grad);

    /* gradient of [outputW, outputB] */
    _MatrixMul(hidden, X_TRANS, gradIn, X_NOTRANS, outputW.grad);
    int shapeSum[]{outputB.GetSize()};
    outputB.grad->Reshape(1, shapeSum);
    _ReduceSum(gradIn, outputB.grad, 0);
    int shapeRaw[]{outputB.GetSize(), 1};
    outputB.grad->Reshape(2, shapeRaw);

    /* gradient of [input, hiddenW, hiddenB] */
    _HardTanHBackward(hidden, hidden);
    _MultiplyMe(hidden->grad, hidden);
    _MatrixMul(input, X_TRANS, hidden->grad, X_NOTRANS, hiddenW.grad);
    _MatrixMul(hidden->grad, X_NOTRANS, &hiddenW, X_TRANS, gradOut);
    shapeSum[0] = hiddenB.GetSize();
    hiddenB.grad->Reshape(1, shapeSum);
    _ReduceSum(hidden->grad, hiddenB.grad, 0);
    shapeRaw[0] = hiddenB.GetSize();
    hiddenB.grad->Reshape(2, shapeRaw);

    //_ScaleAndShiftMe(hiddenB.grad, 0.01);

    OK;
}

ProjectLayer::~ProjectLayer()
= default;
} // namespace ner