/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northeastern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com)
 */

#include <numeric>

#include "SLTKTrainer.h"
#include "SLTKUtility.h"
#include "../../tensor/XUtility.h"

namespace ner
{

/* train the model */
void Trainer::Train(BiLSTMCRFModel *model) const
{

    Word2Vec w2v(config->embFile.c_str(), config->charDim);
    BatchManager testData(config, "test");
    BatchManager trainData(config, "train");

    LOG("start training...");
    LOG("max epoch: %d", config->maxEpoch);

    int globalStep(0);
    vector<DTYPE> loss;
    clock_t start = clock();

    for(int epoch = 1; epoch < config->maxEpoch + 1; ++epoch){
        for(int iteration = 0; iteration < int(trainData.allSentences.size()); ++iteration){
            globalStep++;

            /* forward-propagation in a mini-batch */
            loss.push_back(model->Forward(w2v, trainData.realLengths.at(iteration),
                           trainData.allTags.at(iteration), trainData.allSentences.at(iteration)));

            /*
             * todo: [optimize] use a decay learning rate
            */
            LOG("step %d loss: %f", globalStep, loss.back());
            const bool doUpdate = (!IsNAN(loss.back()) && !IsINF(loss.back()) && loss.back() < 1e3F);
            if(doUpdate){

                /* back-propagation */
                model->Backward(trainData.realLengths.at(iteration), trainData.allTags.at(iteration));

                /* update the parameters */
                model->Update(config->learningRate);
                if(loss.back() < config->record.loss){
                    config->record.loss = loss.back();
                }
                if(globalStep % config->stepsCheck == 0){
                    const DTYPE avgLoss = accumulate(loss.begin(), loss.end(), 0.0) / loss.size();
                    LOG("epoch: %d, iteration: %d/%u, NER loss: %f, duration: %ld ms", epoch, iteration,
                        static_cast<unsigned>(trainData.allSentences.size()), avgLoss, (clock() - start));
                    loss.clear();
                    start = clock();
                }
            }
            else{
                LOG("unexpected loss: %f", loss.back());
                exit(-1);
            }
        }

        /* save the model if the f1 score on the test set increased */
        if(model->Evaluate("dev", w2v)){
            LOG("new best f1 score on %s: %.3f", "dev", config->record.f1[1]);
            if(model->Evaluate("test", w2v)){
                LOG("new best f1 score on %s: %.3f", "test", config->record.f1[2]);
                model->StoreModel(config->ckptPath.c_str());
            }
        }
        config->MakeRecord();
        if(config->record.f1[1] == 100.0F || config->record.f1[2] == 100.0F){
            LOG("Finish training!");
            return;
        }
    }
    OK;
}

/*
update the model by delta rule
\theta_new = \theta - \lrate * grad
where
\lrate = d^-0.5 * min(stepNum^-0.5, stepNum * warmupStepNum^-1.5)
>> model - the t2t model
>> lr - learning rate
*/
//void Trainer::Update(BiLSTMCRFModel * model, const DTYPE lr)
//{
//    XList params(27);
//
//    model->GetParams(params);
//
//    for(int i = 0; i < params.count; i++){
//        XTensor * para = (XTensor*) params.Get(i);
//        XTensor * paraGrad = para->grad;
//
//        if(para == NULL || paraGrad == NULL)
//            continue;
//
//        CheckNTErrors(para != NULL, "NULL parameter tensor!");
//        CheckNTErrors(paraGrad != NULL, "NULL gradient tensor!");
//
//        if(useAdam){
//            adamBeta1T *= adamBeta1;
//            adamBeta2T *= adamBeta2;
//            DTYPE e = lr * (DTYPE) sqrt(1 - adamBeta2T) / (1 - adamBeta1T);
//            DTYPE d = adamDelta * (DTYPE) sqrt(1 - adamBeta2T);
//
//            /* m = beta_1 * m + (1-beta_1) * grad */
//            XTensor * m = (XTensor*) moments.Get(i);
//            _ScaleAndShiftMe(m, adamBeta1, 0);
//            _Sum(m, paraGrad, m, (1.0F - adamBeta1));
//
//            /* v = beta_2 * v + (1-beta_2) * grad * grad*/
//            XTensor * v = (XTensor*) moments2nd.Get(i);
//            _Multiply(paraGrad, paraGrad, v, adamBeta2 / (1.0F - adamBeta2));
//            _ScaleAndShiftMe(v, (1.0F - adamBeta2), 0);
//
//            /* v2 = m / (sqrt(v) + delta) */
//            XTensor * v2 = NewTensorBuf(v, v->devID, v->mem);
//            _Power(v, v2, 0.5F);
//            _ScaleAndShiftMe(v2, 1.0F, d);
//            _Div(m, v2, v2);
//
//            /* the delta rule */
//            _Sum(para, v2, para, -e);
//
//            DelTensorBuf(v2);
//
//        }
//        else{
//            /* the delta rule */
//            _Sum(para, paraGrad, para, -lr);
//        }
//
//        /* clear gradient */
//        paraGrad->SetZeroAll();
//    }
//}

/*
prepare model for training
>> model - the model for training
*/
void Trainer::PrepareModel(BiLSTMCRFModel * model)
{
    CALL;
    moments.Clear();
    moments2nd.Clear();

    XList ws(27);

    model->GetParams(ws);

    for(int i = 0; i < ws.count; i++){
        XTensor * para = static_cast<XTensor*>(ws.Get(i));

        if(config->useAdam){
            XTensor * m = DBG_NEW XTensor(para);
            XTensor * m2 = DBG_NEW XTensor(para);
            m->SetZeroAll();
            m2->SetZeroAll();
            moments.Add(m);
            moments2nd.Add(m2);
        }
    }

    config->adamBeta1T = 1.0F;
    config->adamBeta2T = 1.0F;
    OK;
}


void NERMain(int argc, const char **argv)
{
    CALL;

    Config config("./data/config.txt");
    if(config.isClean){
        remove(config.ckptPath.c_str());
        config.record.loss = 1e3;
        for(auto &score : {config.record.f1,config.record.precision,config.record.recall,config.record.accuracy}){
            score[0] = 0.0;
            score[1] = 0.0;
            score[2] = 0.0;
        }
        LOG("clean the model");
    }

    Trainer trainer(&config);
    BiLSTMCRFModel *model = DBG_NEW BiLSTMCRFModel(&config);

    /* train the model */
    if(config.isTrain)
        trainer.Train(model);

    /* label the specified file */
    else{
        Word2Vec w2V(config.embFile.c_str(), config.charDim);
        model->Evaluate("test", w2V);
    }
    delete model;
    OK;
}

} // namespace ner